import os
import shutil
from typing import Optional

import click

import writer.serve
from writer import VERSION, wf_project
from writer.deploy import cloud, deploy

CONTEXT_SETTINGS = {'help_option_names': ['-h', '--help']}
@click.group(
    context_settings=CONTEXT_SETTINGS,
    help="Writer Framework CLI",
)
@click.version_option(None, '--version', '-v')
def main():
    pass

@main.command()
@click.option('--host', default="127.0.0.1", help="Host to run the app on")
@click.option('--port', default=None, help="Port to run the app on")
@click.option("--enable-jobs-api", help="Set this flag to enable the Jobs API, allowing you to execute jobs without user interaction.", is_flag=True)
@click.argument('path')
def run(path: str, host: str, port: Optional[int], enable_jobs_api: bool):
    """Run the app from PATH folder in run mode."""

    abs_path = os.path.abspath(path)
    if wf_project.is_project(path) is False and \
        wf_project.can_create_project(path) is True:
        raise click.ClickException(f"There's no Writer Framework project at this location, create a new one with `writer create {path}`")

    if wf_project.is_project(path) is False and \
        wf_project.can_create_project(path) is False:
        raise click.ClickException(f"There's no Writer Framework project at this location : {abs_path}")

    writer.serve.serve(
        abs_path, mode="run", port=port, host=host, enable_server_setup=True, enable_jobs_api=enable_jobs_api)

@main.command()
@click.option('--host', default="127.0.0.1", help="Host to run the app on")
@click.option('--port', default=None, help="Port to run the app on")
@click.option('--enable-remote-edit', help="Set this flag to allow non-local requests in edit mode.", is_flag=True)
@click.option('--enable-server-setup', help="Set this flag to enable server setup hook in edit mode.", is_flag=True)
@click.option("--no-interactive", help="Set this flag to run the app without asking anything to the user.", is_flag=True)
@click.option("--enable-jobs-api", help="Set this flag to enable the Jobs API, allowing you to execute jobs without user interaction.", is_flag=True)
@click.argument('path')
def edit(
    path: str,
    port: Optional[int],
    host: str,
    enable_remote_edit: bool,
    enable_server_setup: bool,
    no_interactive: bool,
    enable_jobs_api: bool,
):
    """Run the app from PATH folder in edit mode."""
    abs_path = os.path.abspath(path)
    if wf_project.is_project(path) is False and \
        wf_project.can_create_project(path) is True and \
        no_interactive is False:
        click.confirm("There’s no Writer Framework project at this location, would you like to create a new one ?", default=False, abort=True)
        create_app(path, template_name="default", overwrite=False)

    if wf_project.is_project(path) is False and \
        wf_project.can_create_project(path) is True:
        raise click.ClickException(f"There’s no Writer Framework project at this location, create a new one with `writer create {path}`")

    if wf_project.is_project(path) is False and \
        wf_project.can_create_project(path) is False:
        raise click.ClickException(f"There’s no Writer Framework project at this location : {abs_path}")

    writer.serve.serve(
        abs_path, mode="edit", port=port, host=host,
        enable_remote_edit=enable_remote_edit, enable_server_setup=enable_server_setup, enable_jobs_api=enable_jobs_api)

@main.command()
@click.argument('path')
@click.option('--template', help="The template to use when creating a new app.")
def create(path, template):
    """Create a new app in PATH folder."""

    abs_path = os.path.abspath(path)
    if os.path.isfile(abs_path):
        raise click.ClickException("A target folder is required to create a Writer Framework app. For example: writer create my_app")

    create_app(os.path.abspath(path), template_name=template)

@main.command()
@click.option('--host', default="127.0.0.1", help="Host to run the app on")
@click.option('--port', default=None, help="Port to run the app on")
@click.option('--enable-remote-edit', help="Set this flag to allow non-local requests in edit mode.", is_flag=True)
def hello(port: Optional[int], host: str, enable_remote_edit):
    """Create and run an onboarding 'Hello' app."""
    create_app("hello", template_name="hello", overwrite=True)
    writer.serve.serve("hello", mode="edit",
       port=port, host=host, enable_remote_edit=enable_remote_edit,
       enable_server_setup=False)


main.add_command(deploy, name="deploy")
main.add_command(cloud)

def create_app(app_path: str, template_name: Optional[str], overwrite=False):
    if template_name is None:
        template_name = "default"

    if wf_project.can_create_project(path=app_path) is False:
        raise click.ClickException("The target folder must be empty or not already exist.")

    server_path = os.path.dirname(__file__)
    template_path = os.path.join(server_path, "app_templates", template_name)

    if not os.path.exists(template_path):
        raise click.ClickException(f"Template { template_name } couldn't be found.")

    shutil.copytree(template_path, app_path, dirs_exist_ok=True)
    # create/update requirements.txt and add writer to it
    requirements_path = os.path.join(app_path, "requirements.txt")
    with open(requirements_path, "a") as f:
        f.write(f"writer=={VERSION}\n")


if __name__ == "__main__":
    main()



# BASE PROMPTS (TAILORED TO EACH COMPANY)

base_prompts = {}

base_prompts["Testco"] = """
System Instruction: Create enticing product description pages for Saturn Snacking products to be featured on the Testco supermarket website in the UK. Utilize the provided data fields to ensure each description is comprehensive and tailored to the Saturn Snacking brand. The aim is to make these product offerings more attractive to customers, thereby increasing purchases.

Prompt Details: Use the provided input data: 

{title}
{description}

to generate detailed and enticing product descriptions. 

"""

base_prompts["SuperEats"] = """
System Instruction: Create enticing product description pages for Saturn Snacking products to be featured on the Super Eats platform, which is an online platform usually accessed via mobile phones. Utilize the provided data fields to ensure each description is comprehensive and tailored to the Saturn Snacking brand. The aim is to make these product offerings more attractive to customers, thereby increasing purchases.

The description should be engaging, informative, and limited to a maximum of three sentences. This brief format should be suitable for customers browsing on the Super Eats platform, providing them with enough information to make an informed choice quickly.

Prompt Details: Use the provided input data: 

{title}
{description}

to generate one compact and enticing product description. 

"""

base_prompts["Midl"] = """
System Instruction: Create enticing product description pages for Saturn Snacking products to be featured on the Midl discount supermarket from Germany. Utilize the provided data fields to ensure each description is comprehensive and tailored to the Saturn Snacking brand. The aim is to make these product offerings more attractive to customers, thereby increasing purchases.

Prompt Details: Use the provided input data: 

{title}
{description}

to generate compact and enticing product descriptions. 

"""

### SEO KEYWORD PROMPT
seo_keywords = """
System Instruction: Analyze a group of product descriptions and identify the SEO keywords that are most frequently used. This analysis will help in understanding the common themes and keywords that are essential for optimizing the product descriptions for search engines.

Prompt Details: Examine the following product descriptions and identify the SEO keywords that are most frequently used. 

{descriptions}

Provide Plotly.js graph specification that includes a JSON dictionary of the top 10 keywords and the number of times they appear in the descriptions. The graph should be visually appealing and easy to interpret, highlighting the most relevant keywords. Use #5551ff as the color for the bars in the graph.

"""

# USER PROMPT TEMPLATE
user_prompt = """
System Instruction: Create enticing product description pages for Saturn Snacking products to be featured on the {name} supermarket website. Utilize the provided data fields to ensure each description is comprehensive and tailored to the Saturn Snacking brand. The aim is to make these product offerings more attractive to customers, thereby increasing purchases.

Prompt Details: Use the provided input data: 

{title}
{description}

to generate detailed and enticing product descriptions. Ensure the description is no longer than {character_max} characters.

"""


import json

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate


class ParseJSON(WorkflowBlock):

    @classmethod
    def register(cls, type: str):
        super(ParseJSON, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="workflows_node",
            writer={
                "name": "Parse JSON",
                "description": "Parses a JSON string.",
                "category": "Other",
                "fields": {
                    "plainText": {
                        "name": "Plain text",
                        "type": "Text",
                        "control": "Textarea"
                    },
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "The request was successful.",
                        "style": "success",
                    },
                    "error": {
                        "name": "Error",
                        "description": "The text provided couldn't be parsed.",
                        "style": "error",
                    },
                },
            }
        ))

    def run(self):
        try:
            plain_text = self._get_field("plainText")
            self.result = json.loads(plain_text)
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e


from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate


class WriterChat(WorkflowBlock):

    @classmethod
    def register(cls, type: str):
        super(WriterChat, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="workflows_node",
            writer={
                "name": "Chat completion",
                "description": "Handles chat completions.",
                "category": "Writer",
                "fields": {
                    "conversationStateElement": {
                        "name": "Conversation state element",
                        "desc": "Where the conversation will be stored",
                        "type": "Text",
                    },
                    "useStreaming": {
                        "name": "Use streaming",
                        "type": "Text",
                        "default": "yes",
                        "options": {
                            "yes": "Yes",
                            "no": "No"
                        }
                    },
                    "tools": {
                        "name": "Tools",
                        "type": "Tools",
                        "default": "{}",
                        "category": "Tools"
                    }
                },
                "outs": {
                    "tools": {
                        "name": "Tools",
                        "field": "tools",
                        "description": "Run associated tools.",
                        "style": "dynamic"
                    },
                    "success": {
                        "name": "Success",
                        "description": "If the function doesn't raise an Exception.",
                        "style": "success",
                    },
                    "error": {
                        "name": "Error",
                        "description": "If the function raises an Exception.",
                        "style": "error",
                    },
                },
            }
        ))

    def _make_callable(self, tool_name: str):
        def callable(**args):
            expanded_execution_environment = self.execution_environment | args
            return_value = self.runner.run_branch(self.component_id, f"tools_{tool_name}", expanded_execution_environment, f"Workflow branch execution (chat tool {tool_name})")

            if return_value is None:
                self.outcome = "error"
                raise ValueError(f'No value has been returned for the outcome branch "{tool_name}". Use the block "Return value" to specify one.')

            return return_value
        return callable

    def run(self):
        try:
            import writer.ai

            conversation_state_element = self._get_field("conversationStateElement", required=True)
            use_streaming = self._get_field("useStreaming", False, "yes") == "yes"
            tools_raw = self._get_field("tools", True)
            tools = []

            for tool_name, tool_raw in tools_raw.items():
                tool_type = tool_raw.get("type")
                tool = None
                if tool_type == "function":
                    tool = writer.ai.FunctionTool(
                        type="function",
                        name=tool_name,
                        description=tool_raw.get("description"),
                        callable=self._make_callable(tool_name),
                        parameters=tool_raw.get("parameters")
                    )
                elif tool_type == "graph":
                    tool = writer.ai.GraphTool(
                        type="graph",
                        graph_ids=tool_raw.get("graph_ids"),
                        subqueries=False,
                        description=tool_name,
                    )
                else:
                    continue
                tools.append(tool)

            conversation = self.evaluator.evaluate_expression(conversation_state_element, self.instance_path, self.execution_environment)

            if conversation is None or not isinstance(conversation, writer.ai.Conversation):
                raise ValueError("The state element specified doesn't contain a conversation. Initialize one using the block 'Initialize chat'.")

            msg = ""
            if not use_streaming:
                msg = conversation.complete(tools=tools)
                conversation += msg
                self._set_state(conversation_state_element, conversation)
            else:
                for chunk in conversation.stream_complete(tools=tools):
                    if chunk.get("content") is None:
                        chunk["content"] = ""
                    msg += chunk.get("content")
                    conversation += chunk
                    self._set_state(conversation_state_element, conversation)
            
            self.result = msg
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e


from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate


class ForEach(WorkflowBlock):
    @classmethod
    def register(cls, type: str):
        super(ForEach, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="workflows_node",
                writer={
                    "name": "For-each loop",
                    "description": "Executes a workflow repeatedly, based on the items provided.",
                    "category": "Logic",
                    "fields": {
                        "workflowKey": {
                            "name": "Workflow Key",
                            "desc": "The workflow which will be executed for each item.",
                            "type": "Workflow Key",
                        },
                        "items": {
                            "name": "Items",
                            "desc": "The item value will be passed in the execution environment and will be available at @{payload.item}, its id at @{payload.itemId}.",
                            "default": "{}",
                            "init": '{ "fr": "France", "pl": "Poland" }',
                            "type": "Object",
                            "control": "Textarea",
                        },
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "The workflow executed successfully.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "The workflow wasn't executed successfully.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            workflow_key = self._get_field("workflowKey")
            items = self._get_field("items", as_json=True)
            base_execution_environment = self.execution_environment

            if not isinstance(items, (list, dict)):
                raise ValueError("Items must be a list or dictionary.")

            if isinstance(items, list):
                workflow_environments = [
                    base_execution_environment | {"payload": {"itemId": i, "item": item}}
                    for i, item in enumerate(items)
                ]
                results = self.runner.run_workflow_pool(workflow_key, workflow_environments)
                self.result = results  # Return as a list

            elif isinstance(items, dict):
                workflow_environments = {
                    str(item_id): base_execution_environment
                    | {"payload": {"itemId": str(item_id), "item": item}}
                    for item_id, item in items.items()
                }
                results = self.runner.run_workflow_pool(
                    workflow_key, list(workflow_environments.values())
                )
                self.result = {
                    item_id: results[i] for i, item_id in enumerate(workflow_environments.keys())
                }

            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



import writer as wf
import writer.ai

# Welcome to Writer Framework! 
# This template is a starting point for your AI apps.
# More documentation is available at https://dev.writer.com/framework


# Initialise the state
wf.init_state({
    "my_app": {
        "title": "AI STARTER"
    },
})


import asyncio
import dataclasses
import logging
import os.path
import time
from abc import ABCMeta, abstractmethod
from typing import Callable, Dict, Optional
from urllib.parse import urlparse

from authlib.integrations.requests_client.oauth2_session import OAuth2Session  # type: ignore
from fastapi import Request, Response, status
from fastapi.responses import HTMLResponse, RedirectResponse
from fastapi.templating import Jinja2Templates

import writer.serve
from writer.core import session_manager
from writer.serve import WriterFastAPI
from writer.ss_types import InitSessionRequestPayload

logger = logging.getLogger('writer')

# Dictionary for storing failed attempts {ip_address: timestamp}
failed_attempts: Dict[str, float] = {}

class Unauthorized(Exception):
    """
    This exception allows you to reject the authentication of a user.

    >>>
    """
    def __init__(self, status_code = 401, message = "Unauthorized", more_info = ""):
        self.status_code = status_code
        self.message = message
        self.more_info = more_info


class Auth:
    """
    Interface to implement authentication in Writer Framework.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def register(self,
                 asgi_app: WriterFastAPI,
                 callback: Optional[Callable[[Request, str, dict], None]] = None,
                 unauthorized_action: Optional[Callable[[Request, Unauthorized], Response]] = None
    ):
        raise NotImplementedError

@dataclasses.dataclass
class BasicAuth(Auth):
    """
    Configure Writer Framework to use Basic Authentication. If this is set, Writer Framework will
    ask anonymous users to authenticate using Basic Authentication.

    >>> _auth = auth.BasicAuth(
    >>>     login=os.getenv('LOGIN'),
    >>>     password=os.getenv('PASSWORD')
    >>> )
    >>> writer.server.register_auth(_auth)

    Brute force protection
    ----------------------

    A simple brute force protection is implemented by default. If a user fails to log in, the IP of this user is blocked.
    Writer framework will ban the IP from either the `X-Forwarded-For` header or the `X-Real-IP` header or the client IP address.

    When a user fails to log in, they wait 1 second before they can try again. This time can be modified by
    modifying the value of `delay_after_failure`.

    >>> _auth = auth.BasicAuth(
    >>>     login=os.getenv('LOGIN'),
    >>>     password=os.getenv('PASSWORD')
    >>>     delay_after_failure=5 # 5 seconds delay after a failed login
    >>> )
    >>> writer.server.register_auth(_auth)

    The user is stuck by default after a failure.

    >>> _auth = auth.BasicAuth(
    >>>     login=os.getenv('LOGIN'),
    >>>     password=os.getenv('PASSWORD'),
    >>>     delay_after_failure=5,
    >>>     block_webserver_after_failure=False
    >>> )
    """
    login: str
    password: str
    delay_after_failure: int = 1  # limit attempt when authentication fail (reduce brute force risk)
    block_user_after_failure: bool = True  # delay the answer to the user after a failed login

    callback_func: Optional[Callable[[Request, str, dict], None]] = None  # Callback to validate user authentication
    unauthorized_action: Optional[Callable[[Request, Unauthorized], Response]] = None  # Callback to build its own page when a user is not allowed


    def register(self,
                 asgi_app: WriterFastAPI,
                 callback: Optional[Callable[[Request, str, dict], None]] = None,
                 unauthorized_action: Optional[Callable[[Request, Unauthorized], Response]] = None):

        self.unauthorized_action = unauthorized_action
        self.callback_func = callback

        @asgi_app.middleware("http")
        async def basicauth_middleware(request: Request, call_next):
            import base64
            client_ip = _client_ip(request)

            try:
                if client_ip in failed_attempts and time.time() - failed_attempts[client_ip] < self.delay_after_failure:
                    remaining_time = int(self.delay_after_failure - (time.time() - failed_attempts[client_ip]))
                    raise Unauthorized(status_code=429, message="Too Many Requests", more_info=f"You can try to log in every {self.delay_after_failure}s. Your next try is in {remaining_time}s.")

                session_id = session_manager.generate_session_id()
                _auth = request.headers.get('Authorization')
                if _auth is None:
                    return HTMLResponse("", status.HTTP_401_UNAUTHORIZED, {"WWW-Authenticate": "Basic"})

                scheme, data = (_auth or ' ').split(' ', 1)
                if scheme != 'Basic':
                    return HTMLResponse("", status.HTTP_401_UNAUTHORIZED, {"WWW-Authenticate": "Basic"})

                username, password = base64.b64decode(data).decode().split(':', 1)
                if self.callback_func:
                    self.callback_func(request, session_id, {'username': username})
                else:
                    if username != self.login or password != self.password:
                        raise Unauthorized()

                return await call_next(request)
            except Unauthorized as exc:
                if exc.status_code != 429:
                    failed_attempts[client_ip] = time.time()

                    if self.block_user_after_failure:
                        await asyncio.sleep(self.delay_after_failure)

                if self.unauthorized_action is not None:
                    return self.unauthorized_action(request, exc)
                else:
                    templates = Jinja2Templates(directory=os.path.join(os.path.dirname(__file__), "templates"))
                    return templates.TemplateResponse(request=request, name="auth_unauthorized.html", status_code=exc.status_code, context={
                        "status_code": exc.status_code,
                        "message": exc.message,
                        "more_info": exc.more_info
                    })

@dataclasses.dataclass
class Oidc(Auth):
    """
    Configure Writer Framework to use OpenID Connect. If this is set, Writer Framework will
    redirect anonymous users to OpenID Connect issuer.

    The issuer will then
    authenticate the user and redirect back to the Writer Framework application with
    an authorization code. The Writer Framework application will then exchange the
    authorization code for an access token and use the access token to
    authenticate the user and fetch user information.

    >>> oidc = Oidc(
    ...     client_id="xxxxxxx",
    ...     client_secret="xxxxxxxxxxxxx.apps.googleusercontent.com",
    ...     url_authorize="https://accounts.google.com/o/oauth2/auth",
    ...     url_oauthtoken="https://oauth2.googleapis.com/token",
    ...     url_userinfo="https://www.googleapis.com/oauth2/v1/userinfo?alt=json",
    ... )
    >>> writer.server.register_auth(oidc)

    """
    client_id: str
    client_secret: str
    host_url: str
    url_authorize: str
    url_oauthtoken: str
    scope: str = "openid email profile"
    callback_authorize: str = "authorize"
    url_userinfo: Optional[str] = None
    app_static_public: bool = False

    authlib: OAuth2Session = None
    callback_func: Optional[Callable[[Request, str, dict], None]] = None # Callback to validate user authentication
    unauthorized_action: Optional[Callable[[Request, Unauthorized], Response]] = None # Callback to build its own page when a user is not allowed


    def register(self,
                 asgi_app: WriterFastAPI,
                 callback: Optional[Callable[[Request, str, dict], None]] = None,
                 unauthorized_action: Optional[Callable[[Request, Unauthorized], Response]] = None
                 ):

        redirect_url = urljoin(self.host_url, self.callback_authorize)
        host_url_path = urlpath(self.host_url)
        callback_authorize_path = urljoin(host_url_path, self.callback_authorize)

        auth_authorized_prefix_paths = []
        auth_authorized_routes = [callback_authorize_path]

        for asset_path in writer.serve.wf_root_static_assets():
            if asset_path.is_file():
                auth_authorized_routes.append(urljoin(host_url_path, asset_path.name))
            elif asset_path.is_dir():
                auth_authorized_prefix_paths.append(urljoin(host_url_path, asset_path.name))

        if self.app_static_public is True:
            auth_authorized_prefix_paths += [urljoin(host_url_path, "static"), urljoin(host_url_path, "extensions")]

        logger.debug(f"[auth] oidc - url redirect: {redirect_url}")
        logger.debug(f"[auth] oidc - endpoint authorize: {self.url_authorize}")
        logger.debug(f"[auth] oidc - endpoint token: {self.url_oauthtoken}")
        logger.debug(f"[auth] oidc - path: {host_url_path}")
        logger.debug(f"[auth] oidc - auth authorized routes: {auth_authorized_routes}")
        logger.debug(f"[auth] oidc - auth authorized prefix paths: {auth_authorized_prefix_paths}")
        self.authlib = OAuth2Session(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope.split(" "),
            redirect_uri=redirect_url,
            authorization_endpoint=self.url_authorize,
            token_endpoint=self.url_oauthtoken,
        )

        self.unauthorized_action = unauthorized_action
        self.callback_func = callback

        @asgi_app.middleware("http")
        async def oidc_middleware(request: Request, call_next):
            session = request.cookies.get('session')

            is_one_of_url_prefix_allowed = any(request.url.path.startswith(url_prefix) for url_prefix in auth_authorized_prefix_paths)
            if session is not None or request.url.path in auth_authorized_routes or is_one_of_url_prefix_allowed:
                response: Response = await call_next(request)
                return response
            else:
                url = self.authlib.create_authorization_url(self.url_authorize)
                response = RedirectResponse(url=url[0])
                return response

        @asgi_app.get('/' + urlstrip(self.callback_authorize))
        async def route_callback(request: Request):
            self.authlib.fetch_token(url=self.url_oauthtoken, authorization_response=str(request.url))
            try:
                host_url_path = urlpath(self.host_url)
                response = RedirectResponse(url=host_url_path)
                session_id = session_manager.generate_session_id()

                app_runner = writer.serve.app_runner(asgi_app)
                await app_runner.init_session(InitSessionRequestPayload(
                    cookies=request.cookies, headers=request.headers, proposedSessionId=session_id))

                userinfo = {}
                if self.url_userinfo:
                    userinfo = self.authlib.get(self.url_userinfo).json()

                if self.callback_func:
                    self.callback_func(request, session_id, userinfo)

                if self.url_userinfo:
                    app_runner.set_userinfo(session_id=session_id, userinfo=userinfo)

                response.set_cookie(key="session", value=session_id, httponly=True)
                return response
            except Unauthorized as exc:
                if self.unauthorized_action is not None:
                    return self.unauthorized_action(request, exc)
                else:
                    templates = Jinja2Templates(directory=os.path.join(os.path.dirname(__file__), "templates"))
                    return templates.TemplateResponse(request=request, name="auth_unauthorized.html", status_code=exc.status_code, context={
                        "status_code": exc.status_code,
                        "message": exc.message,
                        "more_info": exc.more_info
                    })


def Google(client_id: str, client_secret: str, host_url: str, app_static_public = False) -> Oidc:
    """
    Configure Google Social login configured through Client Id for Web application in Google Cloud Console.

    >>> import writer.auth
    >>> oidc = writer.auth.Google(client_id="xxxxxxx", client_secret="xxxxxxxxxxxxx.apps.googleusercontent.com", host_url="http://localhost:5000")

    :param client_id: client id of Web application
    :param client_secret: client secret of Web application
    :param host_url: The URL of the Writer Framework application (for callback)
    :param app_static_public: authorizes the exposure of the user's static assets (/static and /extensions)
    """
    return Oidc(
        client_id=client_id,
        client_secret=client_secret,
        host_url=host_url,
        url_authorize="https://accounts.google.com/o/oauth2/auth",
        url_oauthtoken="https://oauth2.googleapis.com/token",
        url_userinfo="https://www.googleapis.com/oauth2/v1/userinfo?alt=json",
        app_static_public=app_static_public
    )

def Github(client_id: str, client_secret: str, host_url: str, app_static_public = False) -> Oidc:
    """
    Configure Github authentication.

    >>> import writer.auth
    >>> oidc = writer.auth.Github(client_id="xxxxxxx", client_secret="xxxxxxxxxxxxx", host_url="http://localhost:5000")

    :param client_id: client id
    :param client_secret: client secret
    :param host_url: The URL of the Writer Framework application (for callback)
    :param app_static_public: authorizes the exposure of the user's static assets (/static and /extensions)
    """

    return Oidc(
        client_id=client_id,
        client_secret=client_secret,
        host_url=host_url,
        url_authorize="https://github.com/login/oauth/authorize",
        url_oauthtoken="https://github.com/login/oauth/access_token",
        url_userinfo="https://api.github.com/user",
        app_static_public=app_static_public
    )

def Auth0(client_id: str, client_secret: str, domain: str, host_url: str, app_static_public = False) -> Oidc:
    """
    Configure Auth0 application for authentication.

    >>> import writer.auth
    >>> oidc = writer.auth.Auth0(client_id="xxxxxxx", client_secret="xxxxxxxxxxxxx", domain="xxx-xxxxx.eu.auth0.com", host_url="http://localhost:5000")

    :param client_id: client id
    :param client_secret: client secret
    :param domain: Domain of the Auth0 application
    :param host_url: The URL of the Writer Framework application (for callback)
    :param app_static_public: authorizes the exposure of the user's static assets (/static and /extensions)
    """

    return Oidc(
        client_id=client_id,
        client_secret=client_secret,
        host_url=host_url,
        url_authorize=f"https://{domain}/authorize",
        url_oauthtoken=f"https://{domain}/oauth/token",
        url_userinfo=f"https://{domain}/userinfo",
        app_static_public=app_static_public
    )

def urlpath(url: str):
    """
    >>> urlpath("http://localhost/app1")
    >>> "/app1"

    >>> urlpath("http://localhost")
    >>> "/"
    """
    path = urlparse(url).path
    if len(path) == 0:
        return "/"
    else:
        return path

def urljoin(*args):
    """
    >>> urljoin("http://localhost/app1", "edit")
    >>> "http://localhost/app1/edit"

    >>> urljoin("app1/", "edit")
    >>> "app1/edit"

    >>> urljoin("app1", "edit")
    >>> "app1/edit"

    >>> urljoin("/app1/", "/edit")
    >>> "/app1/edit"
    """
    root_part = args[0]
    root_part_is_root_path = root_part.startswith('/') and len(root_part) > 1

    url_strip_parts = []
    for part in args:
        if part:
            url_strip_parts.append(urlstrip(part))

    return '/'.join(url_strip_parts) if root_part_is_root_path is False else '/' + '/'.join(url_strip_parts)

def urlstrip(url_path: str):
    """

    >>> urlstrip("/app1/")
    >>> "app1"

    >>> urlstrip("http://localhost/app1")
    >>> "http://localhost/app1"

    >>> urlstrip("http://localhost/app1/")
    >>> "http://localhost/app1"
    """
    return url_path.strip('/')

def _client_ip(request: Request) -> str:
    """
    Get the client IP address from the request.

    >>> _client_ip(request)
    """
    x_forwarded_for = request.headers.get("X-Forwarded-For")
    if x_forwarded_for:
        # X-Forwarded-For can contain a list of IPs, the first is the real IP of the client
        ip = x_forwarded_for.split(",")[0].strip()
    else:
        # Otherwise, use the direct connection IP
        ip = request.headers.get("X-Real-IP", request.client.host)  # type: ignore

    return ip



import io
import sys
from typing import Any

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate

INIT_CODE = """
# State is accessible as a global variable. For example:
state["counter"] = 10

# Other variables from the execution environment are also available. For example:
result # Result from the execution of the last block
results # Dictionary with the execution results of each block, with the block id as key
payload # When executing via API or via an UI event with a payload

# To set the output of this block, which will be available via result to the next block:
set_output("a sample result")

"""


class CodeBlock(WorkflowBlock):
    @classmethod
    def register(cls, type: str):
        super(CodeBlock, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="workflows_node",
                writer={
                    "name": "Python code",
                    "description": "Executes Python code.",
                    "category": "Logic",
                    "fields": {
                        "code": {
                            "name": "Code",
                            "type": "Code",
                            "control": "Textarea",
                            "desc": "The code to be executed.",
                            "init": INIT_CODE,
                        },
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "The event handler execution was successful.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "The event handler execution wasn't successful.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def set_output(self, output: Any):
        self.result = output

    def run(self):
        try:
            code = self._get_field("code")
            self.result = None

            writeruserapp = sys.modules.get("writeruserapp")
            block_globals = (
                {
                    "state": self.runner.session.session_state,
                }
                | self.execution_environment
                | writeruserapp.__dict__
                | {"set_output": self.set_output}
            )
            exec(
                code,
                block_globals,
            )

            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



from typing import Any

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate, WriterConfigurationError


class WriterAddToKG(WorkflowBlock):

    @classmethod
    def register(cls, type: str):
        super(WriterAddToKG, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="workflows_node",
            writer={
                "name": "Add to Knowledge Graph",
                "description": "Adds files to an existing knowledge graph.",
                "category": "Writer",
                "fields": {
                    "graphId": {
                        "name": "Graph Id",
                        "type": "Graph Id",
                        "desc": "The id for an existing knowledge graph. It has a UUID format.",
                        "validator": {
                            "type": "string",
                            "format": "uuid",
                        }
                    },
                    "files": {
                        "name": "Files",
                        "type": "Object",
                        "default": "[]",
                        "desc": "A list of files to be uploaded and added to the knowledge graph. You can use files uploaded via the File Input component or specify dictionaries with data, type and name.",
                        "validator": {
                            "type": "array",
                        }
                    },
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "If the execution was successful.",
                        "style": "success",
                    },
                    "error": {
                        "name": "Error",
                        "description": "If the function raises an Exception.",
                        "style": "error",
                    },
                },
            }
        ))

    def _get_prepared_file(self, raw_file: Any):
        if not isinstance(raw_file, dict):
            raise WriterConfigurationError("Files must be dictionaries and contain `data`, `type` and `name` attributes.")

        if "data" not in raw_file or "type" not in raw_file or "name" not in raw_file:
            raise WriterConfigurationError("A file specified as a dictionary must contain `data`, `type` and `name` attributes.")

        return raw_file

    def run(self):
        try:
            import writer.ai

            graph_id = self._get_field("graphId", required=True)
            raw_files = self._get_field("files", as_json=True)
            prepared_files = []

            if not isinstance(raw_files, list):
                raise WriterConfigurationError("Files must be a list.")

            for raw_file in raw_files:
                prepared_files.append(self._get_prepared_file(raw_file))
                
            graph = writer.ai.retrieve_graph(graph_id)
            
            for prepared_file in prepared_files:
                file = writer.ai.upload_file(prepared_file.get("data"),
                                             prepared_file.get("type"),
                                             prepared_file.get("name"))
                graph.add_file(file)

            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



import hashlib
import json
import logging
import os
import threading
import time
from concurrent.futures import Future, ThreadPoolExecutor, wait
from contextlib import contextmanager
from typing import Any, Dict, List, Literal, Optional, Tuple

import writer.blocks
import writer.blocks.base_block
import writer.core
import writer.core_ui
from writer.ss_types import WorkflowExecutionLog, WriterConfigurationError


class WorkflowRunner:
    def __init__(self, session: writer.core.WriterSession):
        self.session = session
        self.executor_lock = threading.Lock()

    @contextmanager
    def _get_executor(self):
        new_executor = None
        try:
            current_app_process = writer.core.get_app_process()
            with self.executor_lock:
                yield current_app_process.executor
        except RuntimeError:
            logging.info(
                "The main pool executor isn't being reused. This is only expected in test or debugging situations."
            )
            new_executor = ThreadPoolExecutor(20)
            yield new_executor  # Pool executor for debugging (running outside of AppProcess)
        finally:
            if new_executor:
                new_executor.shutdown()

    def execute_ui_trigger(
        self, ref_component_id: str, ref_event_type: str, execution_environment: Dict = {}
    ):
        components = self.session.session_component_tree.get_descendents("workflows_root")
        ui_triggers = list(filter(lambda c: c.type == "workflows_uieventtrigger", components))
        for trigger in ui_triggers:
            if trigger.content.get("refComponentId") != ref_component_id:
                continue
            if trigger.content.get("refEventType") != ref_event_type:
                continue
            self.run_branch(trigger.id, "trigger", execution_environment, "UI trigger execution")

    def run_workflow_by_key(self, workflow_key: str, execution_environment: Dict = {}):
        all_components = self.session.session_component_tree.components.values()
        workflows = list(
            filter(
                lambda c: c.type == "workflows_workflow" and c.content.get("key") == workflow_key,
                all_components,
            )
        )
        if len(workflows) == 0:
            raise ValueError(f'Workflow with key "{workflow_key}" not found.')
        workflow = workflows[0]
        return self.run_workflow(
            workflow.id, execution_environment, f"Workflow execution ({workflow_key})"
        )

    def run_workflow_pool(self, workflow_key: str, execution_environments: List[Dict]):
        """
        Executes the same workflow multiple times in parallel with different execution environments.

        :param workflow_key: The workflow identifier (same workflow for all executions).
        :param execution_environments: A list of execution environments, one per execution.
        :return: A list of results in the same order as execution_environments.
        """

        with self._get_executor() as executor:
            futures = [
                executor.submit(self.run_workflow_by_key, workflow_key, env)
                for env in execution_environments
            ]

        wait(futures)  # Important to preserve order, don't switch to as_completed

        results = []
        for future in futures:
            results.append(future.result())

        return results

    def _get_workflow_nodes(self, component_id):
        return self.session.session_component_tree.get_descendents(component_id)

    def _get_branch_nodes(self, base_component_id: str, base_outcome: Optional[str] = None):
        root_node = self.session.session_component_tree.get_component(base_component_id)
        if not root_node:
            raise RuntimeError(
                f'Cannot obtain branch. Could not find component "{base_component_id}".'
            )
        if not root_node:
            return []
        branch_nodes: List[writer.core_ui.Component] = []
        if not root_node.outs:
            return branch_nodes
        for out in root_node.outs:
            if base_outcome is not None and base_outcome != out.get("outId"):
                continue
            branch_root_node = self.session.session_component_tree.get_component(
                out.get("toNodeId")
            )
            if not branch_root_node:
                continue
            branch_nodes.append(branch_root_node)
            branch_nodes += self._get_branch_nodes(out.get("toNodeId"), base_outcome=None)
        return branch_nodes

    def run_branch(
        self,
        base_component_id: str,
        base_outcome: str,
        execution_environment: Dict,
        title: str = "Branch execution",
    ):
        nodes = self._get_branch_nodes(base_component_id, base_outcome)
        return self.run_nodes(nodes, execution_environment, title)

    def run_workflow(
        self, component_id: str, execution_environment: Dict, title="Workflow execution"
    ):
        nodes = self._get_workflow_nodes(component_id)
        return self.run_nodes(nodes, execution_environment, title)

    def _generate_run_id(self):
        timestamp = str(int(time.time() * 1000))
        salt = os.urandom(8).hex()
        raw_id = f"{self.session.session_id}_{timestamp}_{salt}"
        hashed_id = hashlib.sha256(raw_id.encode()).hexdigest()[:24]
        return hashed_id

    def run_nodes(
        self,
        nodes: List[writer.core_ui.Component],
        execution_environment: Dict,
        title: str = "Workflow execution",
    ):
        execution: Dict[str, Optional[writer.blocks.base_block.WorkflowBlock]] = {}
        tool_futures: Dict[str, Future] = {}
        return_value = None
        execution_environment["run_id"] = self._generate_run_id()
        trace = execution_environment.get("trace", []).copy()
        try:
            futures = []
            with self._get_executor() as executor:
                for node in self.get_terminal_nodes(nodes):
                    futures.append(
                        executor.submit(
                            self.run_node,
                            node,
                            nodes,
                            execution_environment,
                            execution,
                            tool_futures,
                            trace,
                        )
                    )

            wait(futures)

            for tool in execution.values():
                if tool and tool.return_value is not None:
                    return_value = tool.return_value
        except WriterConfigurationError:
            self._generate_run_log(
                execution,
                title,
                "error",
                msg="Execution finished.",
                run_id=execution_environment.get("run_id"),
            )
            # No need to re-raise, it's a configuration error under control and shown as message in the relevant tool
        except BaseException as e:
            self._generate_run_log(
                execution,
                title,
                "error",
                msg="Execution finished.",
                run_id=execution_environment.get("run_id"),
            )
            raise e
        else:
            self._generate_run_log(
                execution,
                "Workflow execution finished",
                "info",
                return_value,
                msg="Execution finished.",
                run_id=execution_environment.get("run_id"),
            )
        return return_value

    def _summarize_data_for_log(self, data):
        if isinstance(data, list):
            return [self._summarize_data_for_log(item) for item in data]
        if isinstance(data, dict):
            return {k: self._summarize_data_for_log(v) for k, v in data.items()}
        if isinstance(data, (str, int, float, bool, type(None))):
            return data
        try:
            return json.loads(json.dumps(data))
        except (TypeError, OverflowError):
            return f"Can't be displayed in the log. Value of type: {str(type(data))}."

    def _generate_run_log(
        self,
        execution: Dict[str, Optional[writer.blocks.base_block.WorkflowBlock]],
        title: str,
        entry_type: Literal["info", "error"],
        return_value: Optional[Any] = None,
        msg: str = "",
        run_id: Optional[str] = None,
    ):
        if not writer.core.Config.is_mail_enabled_for_log:
            return
        if run_id is None:
            run_id = self._generate_run_id()
        exec_log: WorkflowExecutionLog = WorkflowExecutionLog(summary=[])
        for component_id, tool in execution.items():
            if tool is None:
                exec_log.summary.append({"componentId": component_id})
                continue
            exec_log.summary.append(
                {
                    "componentId": component_id,
                    "outcome": tool.outcome,
                    "message": tool.message,
                    "result": self._summarize_data_for_log(tool.result),
                    "returnValue": self._summarize_data_for_log(tool.return_value),
                    "executionEnvironment": self._summarize_data_for_log(
                        tool.execution_environment
                    ),
                    "executionTimeInSeconds": tool.execution_time_in_seconds,
                }
            )

        self.session.session_state.add_log_entry(
            entry_type, title, msg, workflow_execution=exec_log, id=run_id
        )

    def get_terminal_nodes(self, nodes):
        return [node for node in nodes if not node.outs]

    def _get_node_dependencies(
        self, target_node: writer.core_ui.Component, nodes: List[writer.core_ui.Component]
    ) -> List[Tuple]:
        dependencies: List[Tuple] = []
        parent_id = target_node.parentId
        if not parent_id:
            return []
        for node in nodes:
            if not node.outs:
                continue
            for out in node.outs:
                to_node_id = out.get("toNodeId")
                out_id = out.get("outId")
                if to_node_id == target_node.id:
                    dependencies.append((node, out_id))
        return dependencies

    def _is_outcome_managed(self, target_node: writer.core_ui.Component, target_out_id: str):
        if not target_node.outs:
            return False
        for out in target_node.outs:
            if out.get("outId") == target_out_id:
                return True
        return False

    def run_node(
        self,
        target_node: writer.core_ui.Component,
        nodes: List[writer.core_ui.Component],
        execution_environment: Dict,
        execution: Dict[str, Optional[writer.blocks.base_block.WorkflowBlock]],
        tool_futures: Dict[str, Future],
        trace: List[str],
    ):
        tool_class = writer.blocks.base_block.block_map.get(target_node.type)
        if not tool_class:
            raise RuntimeError(f'Could not find tool for "{target_node.type}".')
        dependencies = self._get_node_dependencies(target_node, nodes)

        execution[target_node.id] = None
        result = None
        matched_dependencies = 0
        dependencies_futures = []

        with self._get_executor() as executor:
            for node, out_id in dependencies:
                tool_future = tool_futures.get(node.id)

                if tool_future is None:
                    tool_future = executor.submit(
                        self.run_node,
                        node,
                        nodes,
                        execution_environment,
                        execution,
                        tool_futures,
                        trace,
                    )
                    tool_futures[node.id] = tool_future
                dependencies_futures.append(tool_future)

        wait(dependencies_futures)  # Important to preserve order, don't switch to as_completed

        for future, dependency in zip(dependencies_futures, dependencies):
            (node, out_id) = dependency
            tool = future.result()

            if not tool:
                continue

            if tool.outcome == out_id:
                matched_dependencies += 1

            result = tool.result

            if tool.return_value is not None:
                return tool.return_value

        if len(dependencies) > 0 and matched_dependencies == 0:
            return

        trace += [target_node.id]
        expanded_execution_environment = execution_environment | {
            "trace": trace.copy(),
            "result": result,
            "results": {k: v.result if v is not None else None for k, v in execution.items()},
        }
        tool = tool_class(target_node.id, self, expanded_execution_environment)

        try:
            tool.outcome = "in_progress"
            del execution[target_node.id]
            execution[target_node.id] = tool
            self._generate_run_log(
                execution,
                "Running workflow...",
                "info",
                msg="Execution in progress.",
                run_id=execution_environment.get("run_id"),
            )
            start_time = time.time()
            tool.run()
            tool.execution_time_in_seconds = time.time() - start_time
        except BaseException as e:
            if not tool:
                raise e
            if self._is_outcome_managed(target_node, tool.outcome):
                return tool
            if not tool.outcome or tool.outcome == "in_progress":
                tool.outcome = "error"
            if isinstance(e, WriterConfigurationError):
                tool.message = str(e)
            else:
                tool.message = repr(e)
            raise e
        finally:
            self._generate_run_log(
                execution,
                "Running workflow...",
                "info",
                msg="Execution in progress.",
                run_id=execution_environment.get("run_id"),
            )

        return tool



"""
This module manipulates the folder of a wf project stored into `wf`.

>>> wf_project.write_files_async('app/hello', metadata={"writer_version": "0.1" }, components=...)

>>> metadata, components = wf_project.read_files('app/hello')
"""
import dataclasses
import glob
import io
import json
import logging
import multiprocessing
import os
import queue
import time
import typing
from collections import OrderedDict
from multiprocessing import Queue
from typing import Any, Dict, List, Tuple

from writer import core_ui
from writer.ss_types import ComponentDefinition, MetadataDefinition, SourceFilesDirectory

ROOTS = ['root', 'workflows_root']
COMPONENT_ROOTS = ['page', 'workflows_workflow']

shared_queue_write_files: typing.Optional[Queue] = None

@dataclasses.dataclass
class WfProjectContext:
    app_path: str
    write_files_async_queue: Queue = Queue()
    write_files_async_process: typing.Optional[multiprocessing.Process] = None
    write_files_async_stop: Any = multiprocessing.Event() # Note: Event is a function, not a class in python, it can't be typed


def write_files_async(context: WfProjectContext, metadata: MetadataDefinition, components: Dict[str, ComponentDefinition]) -> None:
    """
    This operation is asynchrone. It's managed in wf_project.process_write_files_async.

    see wf_project.write_files for description

    >>> wf_project.write_files_async('app/hello', metadata={"writer_version": "0.1" }, components=...)
    """
    context.write_files_async_queue.put((context.app_path, metadata, components))

def write_files(app_path: str, metadata: MetadataDefinition, components: Dict[str, ComponentDefinition]) -> None:
    """
    Writes the meta data of the WF project to the `.wf` directory (metadata, components, ...).

    * the metadata.json file is written in json format
    * a file for the root component written in jsonline format
    * a file for the workflows_root component written in jsonline format
    * one file per page is created in the form `components-page-{id}.json` in jsonline format
    * one file per workflow is created in the form `components-workflows_workflow-{id}.json` in jsonline format

    >>> wf_project.write_files('app/hello', metadata={"writer_version": "0.1" }, components=...)
    """

    wf_directory = os.path.join(app_path, ".wf")
    if not os.path.exists(wf_directory):
        os.makedirs(wf_directory)

    with io.open(os.path.join(wf_directory, "metadata.json"), "w") as f:
        json.dump(metadata, f, indent=4)

    _write_root_files(wf_directory, components)
    _write_component_files(wf_directory, components)
    _remove_obsolete_component_files(wf_directory, components)


def start_process_write_files_async(context: WfProjectContext, save_interval: float) -> None:
    """
    Creates a process that writes the .wf/ files

    This agent allows you to process the backup of .wf files in the background
    without blocking application requests.

    :param wf_project_save_interval: the interval in seconds to save the project files
    :return:
    """
    p = multiprocessing.Process(
        target=_start_process_write_files_async_process,
        args=(context, save_interval, context.write_files_async_stop)
    )
    context.write_files_async_process = p
    p.start()


def shutdown_process_write_files_async(context: WfProjectContext) -> None:
    """
    Shutdown the process that writes the .wf/ files
    """
    if context.write_files_async_process is not None and \
        context.write_files_async_process.is_alive():
        context.write_files_async_stop.set()
        context.write_files_async_process.join()
        context.write_files_async_process = None
        context.write_files_async_stop.clear()

def read_files(app_path: str) -> Tuple[MetadataDefinition, dict[str, ComponentDefinition]]:
    """
    Reads project files in the `.wf` folder.

    The components are read in page and workflows_workflow order.

    >>> metadata, components = wf_project.read_files('app/hello')
    """
    components: dict[str, ComponentDefinition] = {}
    roots = ['root', 'workflows_root']
    component_part_file_type = ['page', 'workflows_workflow']

    meta_data_path = os.path.join(app_path, ".wf", "metadata.json")
    try:
        with io.open(meta_data_path, "r") as filep:
            metadata: MetadataDefinition = json.load(filep)
    except Exception as e:
        raise ValueError(f"Error reading metadata file {meta_data_path} : {e}")


    # read ui files
    for root in roots:
        root_component_path = os.path.join(app_path, ".wf", f"components-{root}.jsonl")
        if not os.path.exists(root_component_path):
            continue

        try:
            with io.open(root_component_path, "r") as filep:
                root_component: ComponentDefinition = json.loads(filep.read())
                components.update({root_component["id"]: root_component})
        except Exception as e:
            raise ValueError(f"Error reading root component file {root_component_path} : {e}")

    files = sorted(os.listdir(os.path.join(app_path, ".wf")))
    for file in files:
        for part_file_type in component_part_file_type:
            if file.startswith(f"components-{part_file_type}-"):
                page_file_path = os.path.join(app_path, ".wf", file)
                try:
                    with io.open(page_file_path, "r") as filep:
                        for line in filep:
                            component: ComponentDefinition = json.loads(line)
                            components.update({component["id"]: component})
                except Exception as e:
                    raise ValueError(f"Error reading page component file {page_file_path} : {e}")

    return metadata, components


def migrate_obsolete_ui_json(app_path: str, metadata: MetadataDefinition) -> None:
    """
    Migrates a project that uses ui.json file to the current project format

    The ui.json file is removed after the migration.

    >>> wf_project.migrate_obsolete_ui_json('app/hello')
    """
    assert os.path.isfile(os.path.join(app_path, "ui.json")), f"ui.json file required for migration into {app_path}"

    logger = logging.getLogger('writer')
    with io.open(os.path.join(app_path, "ui.json"), "r") as f:
        parsed_file = json.load(f)

    if not isinstance(parsed_file, dict):
        raise ValueError("No dictionary found in components file.")

    file_payload = parsed_file
    # metadata = file_payload.get("metadata", {})
    components = file_payload.get("components", {})
    write_files(app_path, metadata, components)
    os.remove(os.path.join(app_path, "ui.json"))
    logger.warning('project format has changed and has been migrated with success. ui.json file has been removed.')


def create_default_workflows_root(abs_path: str) -> None:
    with io.open(os.path.join(abs_path, '.wf', 'components-workflows_root.jsonl'), 'w') as f:
        f.write('{"id": "workflows_root", "type": "workflows_root", "content": {}, "isCodeManaged": false, "position": 0, "handlers": {}, "visible": {"expression": true, "binding": "", "reversed": false}}')
        logger = logging.getLogger('writer')
        logger.warning('project format has changed and has been migrated with success. components-workflows_root.jsonl has been added.')


def _expected_component_fileinfos(components: dict[str, ComponentDefinition]) -> List[Tuple[str, str]]:
    """
    Returns the list of component file information to write (id, filename).

    >>> component_files = expected_component_files(components)
    >>> for component_id, filename in component_files:
    >>>     print(f"Components that depends on {component_id} will be written in {filename}")
    """
    expected_component_files = []
    for component_file_root in COMPONENT_ROOTS:
        position = 0
        for c in components.values():
            if c["type"] == component_file_root:
                filename = f"components-{ component_file_root }-{position}-{c['id']}.jsonl"
                expected_component_files.append((c["id"], filename))
                position += 1

    return expected_component_files


def _list_component_files(wf_directory: str) -> List[str]:
    """
    List the component files of the .wf folder.

    The component files are of the form `components-{type}-....jsonl`
    """
    patterns = [f"components-{cr}-" for cr in COMPONENT_ROOTS]
    return [f for f in os.listdir(wf_directory) if any(p in f for p in patterns)]


def _remove_obsolete_component_files(wf_directory: str, components: Dict[str, ComponentDefinition]):
    """
    Remove the obsolete component files in the .wf folder.
    """
    expected_component_files = {f for cid, f in _expected_component_fileinfos(components)}
    list_component_files = set(_list_component_files(wf_directory))

    obsolete_components_files = list_component_files - expected_component_files
    for f in obsolete_components_files:
        os.remove(os.path.join(wf_directory, f))


def _sort_wf_component_keys(obj: typing.Union[ComponentDefinition, dict]) -> Any:
    """
    Sorts the keys of the object recursively to have a consistent order in the json file.

    Id and type attributes first, then use alphabetical order.
    """
    if not isinstance(obj, dict):
        return obj

    def sort_key(item):
        return item[0].replace("id", "0_id").replace("type", "1_type")

    sorted_items = sorted(obj.items(), key=sort_key)
    return OrderedDict((k, _sort_wf_component_keys(v)) for k, v in sorted_items)


def _write_component_files(wf_directory: str, components: Dict[str, ComponentDefinition]) -> None:
    """
    Writes the component files in the .wf folder. It preserve obsolete files.

    The components are written in the form `components-{type}-{position}-{id}.jsonl`
    """
    for component_id, filename in _expected_component_fileinfos(components):
        filtered_components = core_ui.filter_components_by(components, parent=component_id)

        with io.open(os.path.join(wf_directory, filename), "w") as f:
            for p in _order_components(filtered_components):
                f.write(json.dumps(_sort_wf_component_keys(p)) + "\n")


def _write_root_files(wf_directory, components):
    for root in ROOTS:
        root_component = components.get(root, None)
        if root_component:
            with io.open(os.path.join(wf_directory, f"components-{root}.jsonl"), "w") as f:
                f.write(json.dumps(_sort_wf_component_keys(root_component)))


def _order_components(components: Dict[str, ComponentDefinition]) -> List[ComponentDefinition]:
    """
    Orders the components by their position attribute

    >>> ordered_components = _order_components(components)
    """

    def _hierarchical_position(components, c):
        p = [c['position']]
        while c['parentId'] is not None and c['parentId'] in components:
            c = components[c['parentId']]
            p.append(c['position'])

        return list(reversed(p))

    return sorted(components.values(), key=lambda c: _hierarchical_position(components, c))


def _start_process_write_files_async_process(context: WfProjectContext, save_interval, stop_event):
    while True:
        if stop_event.is_set():
            break

        try:
            obj = context.write_files_async_queue.get(block=False)
        except queue.Empty:
            obj = None

        if obj is not None:
            app_path, metadata, components = obj
            write_files(app_path, metadata, components)

        time.sleep(save_interval)


def is_project(path: str) -> bool:
    """
    Returns True if the path is a Writer Framework project.

    >>> wf_project.is_project('app/hello')
    """
    has_main_py = os.path.isfile(os.path.join(path, "main.py"))
    has_wf_directory = os.path.isdir(os.path.join(path, ".wf"))
    has_ui_json_file = os.path.isfile(os.path.join(path, "ui.json"))

    return has_main_py and (has_wf_directory or has_ui_json_file)


def can_create_project(path: str) -> bool:
    """
    Returns True the path does not contain a Writer Framework project and
    it is possible to create one automatically.

    >>> wf_project.can_create_project('app/hello')
    """
    if not os.path.isdir(path):
        return True

    if len(os.listdir(path)) == 0:
        return True

    return False


def build_source_files(app_path: str) -> SourceFilesDirectory:
    """
    Build a file tree as `Dict` wherein the key represent the filename. The value is a `Dict` with a `type` as:

    - `directory`, so it's a directory containing `children`
    - `file`, so it's a file with `content` as string. We limit the file to the first X characters and set `"complete": False` if content is truncated

    Example:

    >>> {'type': 'directory', 'children': {'README.md': {'type': 'file', 'content': 'This app w', 'complete': False}}}
    """
    def load_persisted_script(file: str) -> str:
        path = os.path.join(app_path, file)
        with open(path, "r", encoding='utf-8') as f:
            return f.read()


    ignore_dirs = ['.venv', 'venv', '__pycache__', '.wf']
    ignore_files = ['.DS_Store']
    files = []
    for root, dirs, filenames in os.walk(app_path):
        dirs[:] = [d for d in dirs if d not in ignore_dirs]
        for filename in filenames:
            if filename not in ignore_files:
                files.append(os.path.realpath(os.path.join(root, filename)))

    file_tree: SourceFilesDirectory = { "type": "directory", "children": {} }

    for file in files:
        relative_path = os.path.relpath(file, app_path)
        parts = relative_path.split(os.sep)
        current_level: SourceFilesDirectory = file_tree

        for part in parts:
            if os.path.isdir(os.path.join(app_path, *parts[:parts.index(part) + 1])):
                if part not in current_level["children"]:
                    current_level["children"][part] = { "type": "directory", "children": {} }

                next_level = current_level["children"][part]

                if next_level["type"] == "directory":
                    current_level = next_level
            else:
                try:
                    content = load_persisted_script(relative_path)
                    extension = os.path.splitext(relative_path)[1]
                    # limit only the first 100 characters to limit bandwidth usage, the rest will be lazy loaded
                    exerpt = content if extension == '.py' else content[0:100]
                    current_level["children"][part] = {
                        "type": "file",
                        "content": exerpt,
                        "complete": exerpt == content,
                    }
                except UnicodeDecodeError:
                    # TODO: get mimetype
                    current_level["children"][part] = { "type": "binary" }
                except  FileNotFoundError:
                    pass

    return file_tree



from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate


class RunWorkflow(WorkflowBlock):
    @classmethod
    def register(cls, type: str):
        super(RunWorkflow, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="workflows_node",
                writer={
                    "name": "Run blueprint",
                    "description": "Executes a workflow with a given key.",
                    "category": "Logic",
                    "fields": {
                        "workflowKey": {
                            "name": "Workflow Key",
                            "type": "Workflow Key",
                            "validator": {
                                "type": "string",
                                "format": "writer#workflowKey",
                            },
                        },
                        "payload": {
                            "name": "Payload",
                            "desc": "The value specified will be available using the template syntax i.e. @{payload}",
                            "default": "{}",
                            "type": "Text",
                            "control": "Textarea",
                        },
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "The request was successful.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "The workflow was executed successfully.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            workflow_key = self._get_field("workflowKey")
            payload = self._get_field("payload")
            expanded_execution_environment = self.execution_environment | {"payload": payload}
            return_value = self.runner.run_workflow_by_key(
                workflow_key, expanded_execution_environment
            )
            self.result = return_value
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



import re

import requests

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate


class HTTPRequest(WorkflowBlock):

    CONTROL_CHARS = re.compile(r"[\x00-\x1f\x7f]")

    @classmethod
    def register(cls, type: str):
        super(HTTPRequest, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="workflows_node",
            writer={
                "name": "HTTP Request",
                "description": "Executes an HTTP request.",
                "category": "Other",
                "fields": {
                    "method": {
                        "name": "Method",
                        "type": "Text",
                        "options": {
                            "GET": "GET",
                            "POST": "POST",
                            "PUT": "PUT",
                            "PATCH": "PATCH",
                            "DELETE": "DELETE"
                        },
                        "default": "GET",
                        "validator": {
                            "type": "string",
                            "enum": ["GET", "POST", "PUT", "PATCH", "DELETE"],
                        }
                    },
                    "url": {
                        "name": "URL",
                        "type": "Text",
                        "validator": {
                            "type": "string",
                            "format": "uri",
                        }
                    },
                    "headers": {
                        "name": "Headers",
                        "type": "Key-Value",
                        "default": "{}",
                        "validator": {
                            "type": "object",
                            "patternProperties": {
                                "^.*$": {
                                    "type": ["string", "number", "boolean"],
                                },
                            },
                            "additionalProperties": True,
                        }
                    },
                    "body": {
                        "name": "Body",
                        "type": "Text",
                        "control": "Textarea"
                    },
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "The request was successful.",
                        "style": "success",
                    },
                    "responseError": {
                        "name": "Response error",
                        "description": "The connection was established successfully but an error response code was received or the response was invalid.",
                        "style": "error",
                    },
                    "connectionError": {
                        "name": "Connection error",
                        "description": "The connection couldn't be established.",
                        "style": "error",
                    },
                },
            }
        ))

    def _clean_json_string(self, s: str) -> str:

        """ Remove control characters, which aren't tolerated by JSON loads() strict mode, from string."""

        return HTTPRequest.CONTROL_CHARS.sub("", s)

    def run(self):
        import json

        try:
            method = self._get_field("method", False, "GET")
            url = self._get_field("url")
            headers = self._get_field("headers", True)
            body = self._clean_json_string(self._get_field("body"))
            req = requests.request(method, url, headers=headers, data=body)
            
            content_type = req.headers.get("Content-Type")
            is_json = content_type and "application/json" in content_type
            
            self.result = {
                "headers": dict(req.headers),
                "status_code": req.status_code,
                "body": req.json() if is_json else req.text
            }
            if req.ok:
                self.outcome = "success"
            else:
                self.outcome = "responseError"
                raise RuntimeError("HTTP response with code " + str(req.status_code))
        except json.JSONDecodeError as e:
            self.outcome = "responseError"
            raise e
        except BaseException as e:
            if not self.outcome:
                self.outcome = "connectionError"
            raise e


import json

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate


class WriterClassification(WorkflowBlock):

    @classmethod
    def register(cls, type: str):
        super(WriterClassification, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="workflows_node",
            writer={
                "name": "Classification",
                "description": "Classify a text.",
                "category": "Writer",
                "fields": {
                    "text": {
                        "name": "Text",
                        "type": "Text",
                    },
                    "categories": {
                        "name": "Categories",
                        "type": "Key-Value",
                        "default": "{}"
                    },
                    "additionalContext": {
                        "name": "Additional context",
                        "type": "Text",
                        "control": "Textarea"
                    },
                },
                "outs": {
                    "category": {
                        "field": "categories",
                        "style": "dynamic"
                    },
                    "error": {
                        "name": "Error",
                        "description": "If the function raises an Exception.",
                        "style": "error",
                    },
                },
            }
        ))

    def run(self):
        try:
            import writer.ai

            text = self._get_field("text", required=True)
            additional_context = self._get_field("additionalContext")
            categories = self._get_field("categories", as_json=True, required=True)

            config = {}

            prompt = f"""
Classify the text under “CONTENT” into one of the following categories:

{ json.dumps(categories) }

Your output should be only the key and not contain anything else. For example: { " , ".join(list(categories.keys())) }.

Additional context:

{ additional_context }

CONTENT:
------
{ text }
"""
            result = writer.ai.complete(prompt, config).strip()
            self.result = result
            self.outcome = f"category_{result}"
        except BaseException as e:
            self.outcome = "error"
            raise e


from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate, WriterConfigurationError


class WriterAddChatMessage(WorkflowBlock):

    @classmethod
    def register(cls, type: str):
        super(WriterAddChatMessage, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="workflows_node",
            writer={
                "name": "Add chat message",
                "description": "Add a message to a conversation.",
                "category": "Writer",
                "fields": {
                    "conversationStateElement": {
                        "name": "Conversation state element",
                        "desc": "Where the conversation is stored",
                        "type": "Text",
                    },
                    "message": {
                        "name": "Message",
                        "type": "Object",
                        "init": '{ "role": "assistant", "content": "Hello" }',
                        "validator": {
                            "type": "object",
                            "properties": {
                                "role": { "type": "string" },
                                "content": { "type": "string" },
                            },
                            "additionalProperties": False,
                        }
                    }
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "If the function doesn't raise an Exception.",
                        "style": "success",
                    },
                    "error": {
                        "name": "Error",
                        "description": "If the function raises an Exception.",
                        "style": "error",
                    },
                },
            }
        ))

    def run(self):
        try:
            import writer.ai

            conversation_state_element = self._get_field("conversationStateElement", required=True)
            message = self._get_field("message", as_json=True, required=True)

            conversation = self.evaluator.evaluate_expression(conversation_state_element, self.instance_path, self.execution_environment)

            if conversation is None or not isinstance(conversation, writer.ai.Conversation):
                raise WriterConfigurationError("The state element specified doesn't contain a conversation. Initialize one using the block 'Initialize chat'.")
            
            writer.ai.Conversation.validate_message(message)
            conversation += message

            self._set_state(conversation_state_element, conversation)            
            self.result = None
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e

    